require '4e_analysis'

shared_examples_for "all actions" do
	it "should have a name" do
		@action.name.should_not == nil
	end
	
	it "should have an action type" do
		[:standard, :move, :minor].should include(@action.action_type)
	end
	
	it "should have a frequency (import a Frequency module)" do
		@action.should respond_to :frequency
	end
end

describe TestAction do
	it_should_behave_like "all actions"
	before do; @action = TestAction.new end
end

shared_examples_for "all attacks" do	
	it_should_behave_like "all actions"
	
	it "should have an attack roll" do
		@action.attack_roll.should_not == nil
	end		
		
	it "should have a hit effect" do
		@action.hit_effect.should_not == nil
	end
end

describe Attack do
	class TestAttack < Attack
		def initialize(opts={})
			super(nil)
			@attack = opts[:attack]
			@crit_effect = opts[:crit] || 20
			@hit_effect = opts[:hit] || 10
			@miss_effect = opts[:miss] || 5
		end
		
		def name; "!!! TestAttack" end
		def attack_roll; @attack end
		def crit_effect; @crit_effect end
		def hit_effect; @hit_effect end
		def miss_effect; @miss_effect end
	end
	
	before do
		@opponent = TestCreature.new
	end	
	
	it "should use hit effect when creature hit" do
		attack = TestAttack.new(:attack => 15)
		attack.act(@opponent)
		result.should == :hit
	end
	
	it "should use miss effect when creature missed" do
		attack = TestAttack.new(:attack => 5)
		attack.act(@opponent)
		result.should == :miss
	end
	
	it "should crit on 20 if attack hits" do
		attack = TestAttack.new(:attack => 20)
		attack.act(@opponent)
		result.should == :crit
	end
	
	it "should only crit on natural 20" do
		Dice.fix [15] do
			attack = TestAttack.new(:attack => 1.d20+5)
			attack.act(@opponent)
			result.should == :hit
		end
	end
	
	it "should hit on 20 even if attack would otherwise miss" do
		Dice.max do
			attack = TestAttack.new(:attack => 1.d20-20)
			attack.act(@opponent)
			result.should == :hit
		end
	end
	
	it "should miss on 1 even if attack would otherwise hit" do
		Dice.min do
			attack = TestAttack.new(:attack => 1.d20+20)
			attack.act(@opponent)
			result.should == :miss
		end
	end
	
	it "default miss effect should be nil" do
		Attack.new(nil).miss_effect.should == nil
	end
	
	it "should ignore miss effect when nil" do
		attack = TestAttack.new(:attack => 1)
		def attack.miss_effect; nil end
		attack.act(@opponent)
		@opponent.hp.should == 100
	end
	
	it "crit should maximize hit damage by default" do
		attack = Attack.new(nil)
		def attack.hit_effect; 2.d10 end
		attack.crit_effect.should == 20
	end
	
	it "crit should restore random rolls" do
		attack = Attack.new(nil)
		def attack.hit_effect; 2.d10 end
		attack.crit_effect
		Dice.dice_logic.should == RandomDice
	end
	
	def result
		case @opponent.hp
			when 80 then :crit
			when 90 then :hit
			when 95 then :miss
			when 100 then :no_effect
			else @opponent.hp
		end
	end
	
	def expect_hit
		@opponent.hp.should == 90
	end
	
	def expect_miss
		@opponent.hp.should == 95
	end
end

describe AtWillFrequency do
	class TestAtWillPower < TestAction
		include AtWillFrequency
	end
	
	before do
		@power = TestAtWillPower.new
	end
	
	it "should not be expended after use" do
		@power.act(nil)
		@power.should_not be_expended
	end
	
	it "should provide frequency" do
		@power.frequency.should == :at_will
	end
end

describe EncounterFrequency do
	class TestEncounterPower < TestAction
		include EncounterFrequency
	end
	
	before do
		@power = TestEncounterPower.new
	end
	
	it "should be expended after use" do
		@power.should_not be_expended
		@power.act(nil)
		@power.should be_expended
	end
	
	it "should restore after short rest" do
		@power.act(nil)
		@power.short_rest
		@power.should_not be_expended
	end
		
	it "should provide frequency" do
		@power.frequency.should == :encounter
	end
end

describe DailyFrequency do
	class TestDailyPower < TestAction
		include DailyFrequency
	end
	
	before do
		@power = TestDailyPower.new
	end
	
	it "should be expended after use" do
		@power.should_not be_expended
		@power.act(nil)
		@power.should be_expended
	end
	
	it "should not be restored by a short rest" do
		@power.act(nil)
		@power.short_rest
		@power.should be_expended
	end
			
	it "should provide frequency" do
		@power.frequency.should == :daily
	end
end

describe Reliable do
	class TestReliablePower < TestAttack
		include DailyFrequency
		include Reliable
		def attack_roll; @roll || 1 end
		def roll=(value); @roll = value end
		def hit_effect; 1 end
	end
	
	before do
		@power = TestReliablePower.new
	end
	
	it "should not be expended on miss" do
		@power.act(TestCreature.new)
		@power.should_not be_expended
	end
	
	it "should be expended on hit" do
		@power.roll=18
		@power.act(TestCreature.new)
		@power.should be_expended
	end
	
	it "should be expended on crit" do
		@power.roll=20
		@power.act(TestCreature.new)
		@power.should be_expended
	end
	
	it "should report keyword" do
		@power.keywords.should == ["Reliable"]
	end
end