require '4e_analysis'

class Array
	def classes
		classes = []
		self.each do |a_class|
			classes.push a_class.class
		end
		return classes
	end		
end

shared_examples_for "all creatures" do
	it "should have a name" do
		@creature.name.should_not == nil
	end

	it "should have powers" do
		@creature.power_list.should_not == nil
	end
	
	it "should not reinstantiate its actions" do
		@creature.actions.should == @creature.actions
	end
	
	it "should have basic stats" do
		@creature.hp.should > 0
		@creature.surges.should >= 0
		@creature.ac.should > 0
	end
end

describe TestCreature do
	it_should_behave_like "all creatures"
	
	before do
		@creature = TestCreature.new
	end
end

describe Creature do
	it "should convert to string" do
		TestCreature.new.to_s.should == "!!! Test Creature"
	end
end

describe Creature, "hit points" do
	before do
		@creature = TestCreature.new
	end
	
	it "should start with max hit points" do
	 	@creature.hp.should == @creature.max_hp
	end
	
	it "damage should reduce hp" do
		@creature.damage(13)
		@creature.hp.should == @creature.max_hp - 13
		@creature.damage(2)
		@creature.hp.should == @creature.max_hp - 15
	end
		
	it "should be dead at and below zero hit points" do
		@creature.should_not be_dead
		@creature.damage(99)
		@creature.should_not be_dead
		@creature.damage(1)
		@creature.should be_dead
		@creature.damage(50)
		@creature.should be_dead
	end

	it "should be bloodied at and below half damage" do
		creature = TestCreature.new(:max_hp => 20)
		creature.should_not be_bloodied
		creature.damage(9)
		creature.should_not be_bloodied
		creature.damage(1)
		creature.should be_bloodied
		creature.damage(5)
		creature.should be_bloodied

		creature = TestCreature.new(:max_hp => 21)
		creature.damage(10)
		creature.should_not be_bloodied
		creature.damage(1)
		creature.should be_bloodied
	end
	
	it "surge value should be one quarter maximum hit points" do
		creature = TestCreature.new(:max_hp => 100)
		creature.surge_value.should == 25
		creature = TestCreature.new(:max_hp => 10)
		creature.surge_value.should == 2
	end
	
	it "should regain hit points when healing" do
		@creature.damage(20)
		@creature.heal(5)
		@creature.hp.should == 85
	end
	
	it "should start from zero when healing negative hit points" do
		@creature.damage(200)
		@creature.heal(5)
		@creature.hp.should == 5
	end
	
	it "should be able to spend healing surges" do
		@creature.damage(60)
		@creature.spend_surge
		@creature.hp.should == 65
		@creature.surges.should == 9
	end
	
	it "should not be able to spend more healing surges than it has" do
		creature = TestCreature.new(:max_surges => 0)
		begin
			creature.spend_surge
			raise "should have been out of surges"
		rescue AssertionError
			# expected exception
		end
	end
	
	it "should not regain more than max hit points when healing" do
		@creature.damage(5)
		@creature.heal(20)
		@creature.hp.should == 100
	end
end

describe Creature, "actions" do
	before do
		@opponent = TestCreature.new
	end

	it "should generate powers from power list" do
		class MyCreature < Creature
			def max_hp; 1 end
			def max_surges; 1 end
			def power_list; [MyAction] end
		end
		class MyAction < Action; end
		creature = MyCreature.new
		creature.powers.classes.should == [MyAction]
	end	

	it "should act on its turn" do
		actions = [TestAction.new(:type => :standard), TestAction.new(:type => :move)]
		creature = TestCreature.new(:powers => actions)

		actions[0].should_not have_acted
		actions[1].should_not have_acted
		creature.take_turn(@opponent)
		actions[0].should have_acted
		actions[1].should have_acted
	end
	
	it "should choose one standard, move, and minor action every turn" do
		standard = TestAction.new(:type => :standard)
		move = TestAction.new(:type => :move)
		minor = TestAction.new(:type => :minor)
		standard2 = TestAction.new(:type => :standard)
		move2 = TestAction.new(:type => :move)
		minor2 = TestAction.new(:type => :minor)
		powers = [standard, standard2, move, move2, minor, minor2]
		creature = TestCreature.new(:powers => powers)
		
		creature.actions.should == [standard, move, minor]
	end
	
	it "should only choose actions that are not expended" do
		standard = TestAction.new(:type => :standard, :expended => true)
		standard2 = TestAction.new(:type => :standard, :expended => false)
		
		creature = TestCreature.new(:powers => [standard, standard2])
		creature.actions.should == [standard2]
	end		
	
	it "should not choose actions with 'do not use' priority" do
		standard = TestAction.new(:priority => :do_not_use)
		standard2 = TestAction.new
		
		creature = TestCreature.new(:powers => [standard, standard2])
		creature.actions.should == [standard2]
	end		
	
	it "should not error if there are no actions of a particular type" do
		standard = TestAction.new(:type => :standard)
		minor = TestAction.new(:type => :minor)
		creature = TestCreature.new(:powers => [standard, minor])
		creature.actions.should == [standard, minor]
	end
end

describe Creature, "defenses" do
	before do
		@creature = TestCreature.new
	end
	
	it "should be hit when attack equals ac" do
		@creature.hit?(9).should == false
		@creature.hit?(10).should == true
		@creature.hit?(11).should == true
	end
end

describe Creature, "short rest" do
	it "should notify powers" do
		action = TestAction.new
		creature = TestCreature.new(:powers => [action])
		def action.short_rest
			@notified = true
		end
		def action.notified?
			@notified
		end
		creature.short_rest
		action.should be_notified
	end		
	
	it "should notify powers that aren't in action list" do
		powers = [TestAction.new, TestAction.new]
		creature = TestCreature.new(:powers => powers)
		powers.each do |power|
			def power.short_rest
				@notified = true
			end
			def power.notified?
				@notified
			end
		end
		creature.short_rest
		powers.each do |power|
			power.should be_notified
		end
	end		
	
	it "should heal, using as many surges as necessary" do
		creature = TestCreature.new(:max_surges => 2)
		creature.damage(90)
		creature.short_rest
		creature.hp.should == 60
		creature.surges.should == 0
	end
	
	it "should not use a surge if more than half would be wasted" do
		creature = TestCreature.new(:max_surges => 1)
		creature.damage(12)
		creature.short_rest
		creature.hp.should == 88
		creature.damage(1)
		creature.short_rest
		creature.hp.should == 100
	end
	
	it "should not spend more surges than it has" do
		creature = TestCreature.new(:max_surges => 0)
		creature.damage(90)
		creature.short_rest
	end
end
