require '4e_analysis'
	
describe Dice do
	it "should support arithmetic" do
		Dice.max do
			(2.d10+5).should == 25
		end
	end
	
	it "should support brutal" do
		Dice.fix [1, 2] do
			1.d10(:brutal => 1).should == 2
		end
	end
	
	it "should represent a single roll" do
		Dice.fix [3] do
			dice = FixedDice.new(1, 4)
			dice.roll.should == 3
			dice.roll.should == 3
		end
	end
	
	it "should provide rolls" do
		Dice.fix [20] do
			1.d20.rolls.should == [20]
		end
	end
	
	it "should support bonuses" do
		dice1 = MaximumDice.new(2, 6, :brutal=>1)
		dice2 = dice1 + 4
		
		dice2.roll.should == 16
	end
	
	it "should support penalties" do
		dice1 = MaximumDice.new(2, 12)
		dice2 = dice1 - 4
		dice2.roll.should == 20
	end
	
	it "should not reroll when appending bonuses" do
		Dice.fix [1] do
			dice = 1.d10
			dice.roll.should == 1
			dice += 1
			dice.roll.should == 2
		end
	end
	
	it "should roll dice before appending bonus" do
		Dice.fix [1] do
			dice = 1.d6
			(dice + 1).should == 2
			(dice + 1).should == 2
		end
	end
		
	it "should chain together bonuses" do
		dice1 = MaximumDice.new(1, 20)
		dice2 = dice1 + 2 - 4 - 6 + 3
		dice2.roll.should == 15
	end

	it "caching roll should work with bonuses" do
		dice = MaximumDice.new(1, 10)+3
		dice.roll.should == 13
		dice.roll.should == 13
	end	

	it "should coerce to integer" do
		dice = MaximumDice.new(1, 10)
		(1 + dice).should == 11
		(2 - dice).should == -8
	end
	
	it "should compare with Fixnum" do
		dice = MaximumDice.new(1, 4)
		(dice < 4).should == false
		(dice <= 4).should == true
		(dice == 4).should == true
		(dice >= 4).should == true
		(dice > 4).should == false
		(dice != 4).should == false
		(4 == dice).should == true
		(4 != dice).should == false
	end
	
	it "should report its formula" do
		MaximumDice.new(1, 10).dice.should == "1d10"
		MaximumDice.new(3, 6).dice.should == "3d6"
		MaximumDice.new(2, 4, :brutal => 2).dice.should == "2d4[brutal 2]"
	end
	
	it "should report details" do
		Dice.fix [3] do
			FixedDice.new(1, 4).details.should == "1d4; rolled 3"
		end
		Dice.fix [2, 5, 6] do
			FixedDice.new(3, 6).details.should == "3d6; rolled 2, 5, 6"
		end
	end
	
	it "should report brutal" do
		Dice.fix [3, 1, 2] do
			FixedDice.new(2, 10, :brutal => 1).details.should == "2d10[brutal 1]; rolled 3, 1, 2"
		end
	end
	
	it "should report bonuses and penalties" do
		Dice.fix [6, 6] do
			dice = 2.d6
			(dice+4).details.should == "2d6+4; rolled 6, 6"
			(dice-4).details.should == "2d6-4; rolled 6, 6"
		end
	end
	
	it "should convert to string" do
		MaximumDice.new(1, 4).to_s.should == "4"
	end
end

describe RandomDice do
	it "should provide a linear distribution" do
		results = []
		1.upto(4) do |i|
			results[i] = 0
		end
		
		1000.times do
			results[RandomDice.new(1, 4).roll] += 1
		end
		
		results[1].should > 200
		results[2].should > 200
		results[3].should > 200
		results[4].should > 200
	end			
end

describe MaximumDice do
	it "should maximize all rolls" do
		MaximumDice.new(1, 10).roll.should == 10
	end
	
	it "should report maximized status" do
		MaximumDice.new(1, 10).details.should == "1d10; maximized"
	end
end

describe MinimumDice do
	it "should minimize all rolls" do
		MinimumDice.new(1, 10).roll.should == 1
		MinimumDice.new(2, 6).roll.should == 2
	end
	
	it "should support brutal" do
		MinimumDice.new(1, 10, :brutal => 3).roll.should == 4
	end
end

describe FixedDice do
	it "should take results from a list" do
		Dice.roll_list = [1, 2, 3, 4]
		FixedDice.new(1, 4).roll.should == 1
		FixedDice.new(1, 4).roll.should == 2
		FixedDice.new(1, 4).roll.should == 3
		FixedDice.new(1, 4).roll.should == 4
	end
	
	it "should raise error when the list is empty" do
		begin
			Dice.roll_list = []
			FixedDice.new(1, 4).roll
			raise "Expected NoMoreDiceError"
		rescue Dice::NoMoreDiceError
			# correct result
		end
	end
				
	it "should raise error when list value is outside possible value" do
		Dice.roll_list = [0, 5]
		begin
			FixedDice.new(1, 4).roll
			raise "Expected ValueTooSmallError"
		rescue Dice::ValueTooSmallError
			# correct result
		end
		
		begin
			FixedDice.new(1, 4).roll
			raise "Expected ValueTooLargeError"
		rescue Dice::ValueTooLargeError
			# correct result
		end
	end
end

describe Dice, "factory" do
	it "should return to random rolls" do
		Dice.min do end
		Dice.dice_logic.should == RandomDice

		Dice.max do end
		Dice.dice_logic.should == RandomDice

		Dice.fix [] do end
		Dice.dice_logic.should == RandomDice
	end
end

describe Dice, "Fixnum extension" do
	it "should roll 1 die" do
		Dice.max do
			1.d4.should == 4
			1.d6.should == 6
			1.d8.should == 8
			1.d10.should == 10
			1.d12.should == 12
			1.d20.should == 20
			1.d100.should == 100
		end
	end
	
	it "should roll multiple dice" do
		Dice.max do
			3.d6.should == 18
			3.d4.should == 12
		end
	end
	
	it "should report 'formula'" do
		3.dice.should == "3"
	end
	
	it "should report 'rolls'" do
		3.rolls.should == [3]
	end
	
	it "should report 'details'" do
		3.details.should == "static"
	end
end
