require '4e_analysis'

class Creature
	attr_reader :hp, :ac, :surges
	attr_writer :report
	
	def initialize(report = Report.new)
		@report = report
		@hp = max_hp
		@surges = max_surges
	end
	
	def report
		@report.puts "#{self}:" do
			report_stats
			@report.puts "Powers:" do
				powers.each do |power|
					power.report
				end
			end				
		end
	end
	
	def report_stats
		@report.puts "Stats: HP #{hp}; AC #{ac}; Surges #{surges}"
	end
	
	def bloodied?
		@hp <= max_hp / 2
	end
	
	def dead?
		@hp <= 0
	end
	
	def surge_value
		max_hp / 4
	end

	def damage(amount)
		original = @hp
		@hp -= amount
		@report.puts "#{self} has #{hp} hp remaining (#{original} hp - #{amount} damage)."
	end
	
	def heal(amount)
		@hp = 0 if @hp < 0
		original = @hp
		@hp += amount
		@hp = max_hp if @hp > max_hp
		wasted = amount - (@hp - original)
		@report.puts "#{self} heals #{amount} hp (#{original} hp -> #{@hp} hp; #{wasted} hp wasted)."
	end
	
	def spend_surge
		@surges -= 1
		@report.puts "#{self} spends a healing surge (#{surges} remaining)."
		heal(surge_value)
		assert @surges >= 0, "Can't spend surges you don't have"
	end
	
	def hit?(attack)
		attack >= ac
	end
		
	def short_rest
		@report.puts "#{self} takes a short rest." do
			min_hp = max_hp - (surge_value / 2)
			while @surges > 0 && @hp < min_hp do
				spend_surge
			end
			powers.each do |power|
				power.short_rest
			end
		end
	end
	
	def take_turn(opponent)
		@report.puts "#{self}'s turn:" do
			report_stats
			actions.each do |action|
				case action.action_type
					when :standard then type = "standard action"
					when :move then type = "move action"
					when :minor then type = "minor action"
					else raise NotImplementedError
				end
				@report.puts "#{action.name} (#{type}):" do
					action.act(opponent)
				end
			end	
		end
	end
	
	def actions
		standard, move, minor = nil, nil, nil
		powers.each do |power|
			next if power.expended? || power.priority == :do_not_use
			standard ||= power if power.action_type == :standard
			move ||= power if power.action_type == :move
			minor ||= power if power.action_type == :minor
		end
		result = []
		result.push standard if standard
		result.push move if move
		result.push minor if minor
		return result
	end
	
	def powers
		unless @powers
			@powers = []
			power_list.each do |power_class|
				@powers.push(power_class.new(self, @report))
			end
		end
		return @powers
	end
	
	def to_s
		name
	end
end

class DeprecatedCreature
	def initialize
		@hp = max_hp
	end

	def dead?
		return @hp <= 0
	end
	
	def bloodied?
		return @hp <= @max_hp / 2
	end
	
	def defend(att, hit_dmg, miss_dmg = 0)
		miss = (att == 1) || (att < @ac)
		@hp -= att >= @ac ? hit_dmg : miss_dmg
		return !miss
	end
end

class TestCreature < Creature
	attr_reader :max_hp, :max_surges, :powers

	def initialize(opts={})
		@max_hp = opts[:max_hp] || 100
		@ac = opts[:ac] || 10
		@max_surges = opts[:max_surges] || 10
		@powers = opts[:powers] || []
		@rounds_to_live = opts[:rounds_to_live]
		@round = 0
		super()
	end
	
	def name
		"!!! Test Creature"
	end
	
	def power_list
		[]
	end
	
	def defend(attack_roll, attack_bonus, attack_action)
		super
		@attacked = true
	end
	
	def take_turn(opponent)
		super
		@had_turn = true
		@round += 1
	end
	
	def dead?
		super || (@rounds_to_live && @round >= @rounds_to_live)
	end
	
	def attacked?
		@attacked
	end
	
	def had_turn?
		@had_turn
	end
end